/*********** 
# Copyright (c) 2017, Xilinx, Inc. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without 
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# 
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software
# without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.
#
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY 
# DIRECT, INDIRECT,INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO,PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY,OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
************/

/*******************************************************************************

    This is a nearest neighbor of a point example which showcases how making 
    a loop-nest perfect or semi-perfect can help in improving performance.

*******************************************************************************/

#include <iostream>
#include <cstdio>
#include <stdlib.h>
#include <xil_cache.h>
#include <xil_io.h>
#include "xparameters.h"
#include "xmutex.h"

#include <cstring>
#include <climits>
#include "nearest.h"
#include "sds_utils.h"

#define GPIO_BASE  0xE000A000
#define GPIO_DIRM  (GPIO_BASE+0x0284)
#define GPIO_MASK_DATA  (GPIO_BASE+0x0010)

/*
 * The following constants map to the XPAR parameters created in the
 * xparameters.h file. They are defined here such that a user can easily
 * change all the needed parameters in one place.
 */
#define MUTEX_DEVICE_ID		XPAR_MUTEX_0_IF_0_DEVICE_ID

/*
 * The Mutex number on which the example is run
 */
#define MUTEX_NUM 		0

// Maximum possible distance between two points
#define INFINITY ULONG_MAX

#define SQUARE(x) ((x)*(x))

/*
 * Mutex instance
 */
XMutex Mutex;

// Software implementation for finding nearest neighbor
void nearest_sw(
                int *in,      // Input Points Array - represented as integer
                int *point,   // Current Point for which the neighbor is found
                int *out,     // Output Point
                int size,     // Size of the input array
                int dim       // #Dimensions of the points
                )
{
    unsigned long curr_dist, min_dist = INFINITY;

    for(int i = 0; i < size; i++) {
        curr_dist = 0;

        // Calculate distance
        // No need to compute square root as the distances are only compared
        for(int j = 0; j < dim; j++) {
            curr_dist += SQUARE(point[j] - in[i*dim + j]);
        }

        // Update minimum distance
        if(curr_dist < min_dist) {

            min_dist = curr_dist;

            for(int k = 0; k < dim; k++) {
                out[k] = in[i*dim + k];
            }
        }
    }
}

int main(int argc, char** argv)
{

	/*
	 * Init mutex and start microblaze
	 */
	XMutex_Config *ConfigPtr;
	XStatus Status;
	u32 TimeoutCount = 0;
	u32 val;

	/*
	 * Lookup configuration data in the device configuration table.
	 * Use this configuration info down below when initializing this
	 * driver instance.
	 */
	ConfigPtr = XMutex_LookupConfig(MUTEX_DEVICE_ID);
	if (ConfigPtr == (XMutex_Config *) NULL) {
		return XST_FAILURE;
	}

	/*
	 * Perform the rest of the initialization.
	 */
	Status = XMutex_CfgInitialize(&Mutex, ConfigPtr, ConfigPtr->BaseAddress);
	if (Status != XST_SUCCESS) {
		return XST_FAILURE;
	}

	XMutex_SetUser(&Mutex, MUTEX_NUM, 0);
	XMutex_Lock(&Mutex, MUTEX_NUM);

#if (1)
	printf("ARMCPU0: place 0xb8000000 at start of MB0 vectors\n\r");
	Xil_Out32(0x28100000, 0xb8000000);
	Xil_DCacheFlush();
#endif

	/*
	 * Take PL out of reset
	 * Enable emio bit 0
	 */
	Xil_Out32(GPIO_DIRM, 0x1);

	/*
	 * Set emio[0] - set PL reset
	 */
	Xil_Out32(GPIO_MASK_DATA, 0xfffe0000);

	/*
	 * Set emio[0] - release PL reset
	 * Microblaze reset released
	 */
	Xil_Out32(GPIO_MASK_DATA, 0xfffe0001);

	printf("ARMCPU0: MB0 reset removed, ARM waiting ...\n\r");

	/*
	 * Wait for response from Microblaze via user value at mutex 0
	 */
	do {
		XMutex_GetUser(&Mutex, MUTEX_NUM, &val);
	} while (val == 0);

	printf("ARMCPU0: MB0 indicates - running ...\n\r");

    if (DATA_DIM > MAX_DIM) {
        std::cout << DATA_DIM << "greater than " << MAX_DIM << "!"<< " Please use a"; 
        std::cout << "smaller DATA_DIM" << std::endl;
        return EXIT_FAILURE;
    }

    if (DATA_SIZE > MAX_SIZE) {
        std::cout << DATA_SIZE << "greater than " << MAX_SIZE << "!"<< " Please use"; 
        std::cout << "a smaller DATA_SIZE" << std::endl;
        return EXIT_FAILURE;
    }

    // Size of input data
    size_t vector_size_bytes = sizeof(int) * DATA_SIZE * DATA_DIM;

    // Allocate buffers using sds_alloc
    int *source_in  =       (int*) sds_alloc(vector_size_bytes);
    int *source_point =     (int*) sds_alloc(sizeof(int)*DATA_DIM);
    int *source_hw_result = (int*) sds_alloc(sizeof(int)*DATA_DIM);
    
    // Allocate software output buffer
    int *source_sw_result = (int*) malloc(sizeof(int)*DATA_DIM);

    // Check for failed memory allocation
    if((source_in == NULL) || (source_point == NULL) || (source_hw_result == NULL) || (source_sw_result == NULL)){
       std::cout << "TEST FAILED : Failed to allocate memory" << std::endl;
       return -1;
     }
   
    // Create the test data
    for(int i = 0 ; i < DATA_SIZE*DATA_DIM; i++){
        source_in[i] = rand()%100;
    }

    for(int i = 0 ; i < DATA_DIM; i++){
        source_point[i] = rand()%100;
    }

    int size = DATA_SIZE;
    int dim = DATA_DIM;

    sds_utils::perf_counter hw_ctr, sw_ctr;

    hw_ctr.start();
    // Launch Hardware Solution
    nearest_accel(source_in, source_point, source_hw_result, size, dim);
    hw_ctr.stop();

    // Launch Software Solution
    nearest_sw(source_in, source_point, source_sw_result, size, dim);

    uint64_t hw_cycles = hw_ctr.avg_cpu_cycles();

    std::cout << "Number of CPU cycles running application in hardware: "
                << hw_cycles << std::endl;

    // Compare the nearest distances between software and hardware
    unsigned long dist_sw = 0, dist_hw = 0;
    for(int i = 0; i < dim; i++) {
        dist_sw += SQUARE(source_sw_result[i] - source_point[i]);
        dist_hw += SQUARE(source_hw_result[i] - source_point[i]);
    }

    // Release Memory 
    sds_free(source_in);
    sds_free(source_point);
    sds_free(source_hw_result);
    free(source_sw_result);

    if(dist_sw != dist_hw)
    {
        std::cout << "TEST FAILED." << std::endl;
        std::cout << "\tSoftware Min Dist = " << dist_sw << std::endl;
        std::cout << "\tHardware Min Dist = " << dist_hw << std::endl;
        return -1;
    }

    std::cout << "TEST PASSED." << std::endl;
    return 0;
}
